import json
import logging
import os
import shlex
import subprocess
from random import getrandbits
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast

from docker.types import LogConfig
from ruamel.yaml import YAMLError

from splitgraph.ingestion.airbyte.docker_utils import (
    detect_network_mode,
    remove_at_end,
    wait_not_failed,
)
from splitgraph.utils.docker import (
    copy_dir_to_container,
    get_docker_client,
    get_file_from_container,
)
from splitgraph.utils.yaml import safe_dump, safe_load

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine


def prepare_git_repo(url: str, target_path: str, ref: str = "master") -> None:
    """
    Clone a repository from Git and check out a specific branch.

    :param url: Git repository URL. If authentication is required, one may use e.g.
        `https://uname:pass_or_token@github.com/organisation/repository.git`
    :param target_path: Path to the repository. Must already exist and be empty.
    :param ref: Branch/commit hash to checkout
    """
    subprocess.check_call(["git", "init"], cwd=target_path)
    subprocess.check_call(["git", "remote", "add", "origin", url], cwd=target_path)
    subprocess.check_call(["git", "fetch", "origin", ref], cwd=target_path)
    subprocess.check_call(["git", "checkout", ref], cwd=target_path)
    logging.info(subprocess.check_output(["git", "log", "--oneline", "-5"], cwd=target_path))


def make_dbt_profile(conn_params: Dict[str, Any], schema: str, threads: int = 32) -> Dict[str, Any]:
    """Build a dbt profile dict (to go into profiles.yml) to output data into our staging schema.

    Note that this doesn't change the project's sources to read data from a schema."""
    return {
        "config": {
            "partial_parse": True,
            "printer_width": 120,
            "send_anonymous_usage_stats": False,
            "use_colors": True,
        },
        "splitgraph": {
            "outputs": {
                "splitgraph": {
                    "type": "postgres",
                    "dbname": conn_params["SG_ENGINE_DB_NAME"],
                    "host": conn_params["SG_ENGINE_HOST"],
                    "port": int(conn_params["SG_ENGINE_PORT"]),
                    "user": conn_params["SG_ENGINE_USER"],
                    "pass": conn_params["SG_ENGINE_PWD"],
                    "schema": schema,
                    "threads": threads,
                }
            },
            "target": "splitgraph",
        },
    }


def patch_dbt_project_sources(
    project_path: str,
    default_schema: str,
    source_schema_map: Optional[Dict[str, str]] = None,
) -> None:
    """
    Patch a dbt project's source definitions to point them to a different schema.

    :param project_path: Path to the dbt project
    :param source_schema_map: Map from dbt source names to the schema name to override
    :param default_schema: Default schema to use
    """

    source_schema_map = source_schema_map or {}

    # This is kind of crude but it doesn't look like dbt lets us override this on the CLI.
    # Basically, we want to make sure the project loads data just from the staging schema
    # (where we've put our LQ mounted foreign tables) as opposed to any schema on the engine.
    #
    # We do this by going through all YAML files in the project and seeing if they have the
    # sources: [...] dbt source spec, then patching the schema in there.
    for dirpath, _, filenames in os.walk(project_path):
        for filename in filenames:
            if not filename.endswith((".yml", ".yaml")):
                continue

            filepath = os.path.join(dirpath, filename)
            logging.debug("Checking %s", filepath)
            with open(filepath, "r") as f:
                try:
                    data = safe_load(f)
                except YAMLError as e:
                    logging.warning("Error loading %s, ignoring", filepath, exc_info=e)
                if not data or "sources" not in data:
                    continue

                for source in data["sources"]:
                    target = source_schema_map.get(source["name"], default_schema)

                    logging.info(
                        "Patching source %s in %s to point to schema %s",
                        source.get("name"),
                        filepath,
                        target,
                    )
                    source["schema"] = target

            with open(filepath, "w") as f:
                safe_dump(data, f)


def run_dbt_transformation_from_git(
    engine: "PsycopgEngine",
    target_schema: str,
    repository_url: str,
    repository_ref: str = "master",
    dbt_image: str = "fishtownanalytics/dbt:1.0.0",
    source_schema_map: Optional[Dict[str, str]] = None,
    default_source_schema: Optional[str] = None,
    models: Optional[List[str]] = None,
) -> None:
    """
    Run a dbt transformation from Git on a dataset checked out into a schema
    :param engine: Engine the data is on
    :param target_schema: The schema to write the dbt-built data out into.
    :param repository_url: URL of the Git repository with the dbt project (including the
        password/access token if required
    :param repository_ref: Branch or commit hash to check out.
    :param dbt_image: Docker image to use for dbt.
    :param source_schema_map: Map of dbt source names to schema names.
    :param default_source_schema: Default source schema for sources not in the map. If not specified,
        `target_schema` is used
    :param models: List of dbt models to limit the build to.
    """

    # Clone the repo into a temporary location and patch it to point to our staging schema
    client = get_docker_client()
    network_mode = detect_network_mode()
    with TemporaryDirectory() as tmp_dir:
        # Prepare the target directory that we'll copy into the dbt container.
        # We can't bind mount it since we ourselves could be running inside of Docker.
        project_path = os.path.join(tmp_dir, "dbt_project")
        os.mkdir(project_path)

        # Add a Git repo and switch all of its sources to use our staging schema
        prepare_git_repo(repository_url, project_path, repository_ref)
        patch_dbt_project_sources(
            project_path,
            source_schema_map=source_schema_map,
            default_schema=default_source_schema or target_schema,
        )

        # Make a dbt profile file that points to our engine
        profile = make_dbt_profile(engine.conn_params, target_schema)
        with open(os.path.join(tmp_dir, "profiles.yml"), "w") as f:
            safe_dump(profile, f)

        # Create the normalization container
        # We'll use Airbyte's image for now (since they have an image with dbt installed and our
        # Airbyte load is currently the only user of this function), but at this point
        # there isn't much that makes us require it.
        entrypoint = ["/bin/bash"]
        command = [
            "-c",
            "dbt deps --project-dir /data/dbt_project && "
            "dbt run --profiles-dir /data --project-dir /data/dbt_project --profile splitgraph",
        ]
        if models is not None:
            command[-1] += " --models " + shlex.join(["+" + m for m in models])
        client.images.pull(dbt_image)
        container = client.containers.create(
            image=dbt_image,
            name="sg-dbt-{:08x}".format(getrandbits(64)),
            entrypoint=entrypoint,
            command=command,
            network_mode=network_mode,
            log_config=LogConfig(type="json-file", config={"max-size": "10m", "max-file": "3"}),
        )

        with remove_at_end(container):
            copy_dir_to_container(container, tmp_dir, "/data", exclude_names=[".git"])
            container.start()
            wait_not_failed(container, mirror_logs=True)


def compile_dbt_manifest(
    engine: "PsycopgEngine",
    repository_url: str,
    repository_ref: str = "master",
    dbt_image: str = "fishtownanalytics/dbt:1.0.0",
) -> Dict[str, Any]:

    # Clone the repo into a temporary location and patch it to point to our staging schema
    client = get_docker_client()
    network_mode = detect_network_mode()
    with TemporaryDirectory() as tmp_dir:
        # Prepare the target directory that we'll copy into the dbt container.
        # We can't bind mount it since we ourselves could be running inside of Docker.
        project_path = os.path.join(tmp_dir, "dbt_project")
        os.mkdir(project_path)

        # Add a Git repo
        prepare_git_repo(repository_url, project_path, repository_ref)

        # Make a dbt profile file that points to our engine
        profile = make_dbt_profile(engine.conn_params, "splitgraph")
        with open(os.path.join(tmp_dir, "profiles.yml"), "w") as f:
            safe_dump(profile, f)

        # Get the location of the target that the build outputs the manifest to
        with open(os.path.join(project_path, "dbt_project.yml"), "r") as f:
            project = safe_load(f)
            build_dir = project.get("target-path", "target")

        entrypoint = ["/bin/bash"]
        command = [
            "-c",
            "dbt deps --project-dir /data/dbt_project && "
            "dbt compile --profiles-dir /data --project-dir /data/dbt_project --profile splitgraph",
        ]
        client.images.pull(dbt_image)
        container = client.containers.create(
            image=dbt_image,
            name="sg-dbt-{:08x}".format(getrandbits(64)),
            entrypoint=entrypoint,
            command=command,
            network_mode=network_mode,
            log_config=LogConfig(type="json-file", config={"max-size": "10m", "max-file": "3"}),
        )

        with remove_at_end(container):
            copy_dir_to_container(container, tmp_dir, "/data", exclude_names=[".git"])
            container.start()
            wait_not_failed(container, mirror_logs=True)
            # Extract the manifest
            return cast(
                Dict[str, Any],
                json.load(
                    get_file_from_container(
                        container, os.path.join("/data/dbt_project/", build_dir, "manifest.json")
                    )
                ),
            )



"""
sgr commands related to getting information out of / about images
"""

from collections import Counter, defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union, cast

import click

from ..core._drawing import format_image_hash, format_tags, format_time
from .common import ImageType, RepositoryType, emit_sql_results, remote_switch_option

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository


@click.command(name="log")
@click.argument("image_spec", type=ImageType(get_image=False, default="HEAD"))
@click.option("-t", "--tree", is_flag=True)
@remote_switch_option()
def log_c(image_spec, tree):
    """
    Show the history of a Splitgraph repository/image.

    By default, this shows the history of the current branch, starting from the HEAD pointer and following its
    parent chain.

    Alternatively, it can follow the parent chain of any other image.

    If ``-t`` or ``--tree`` is passed, this instead renders the full image tree. The repository doesn't need to have
    been checked out in this case.
    """
    from tabulate import tabulate

    from splitgraph.core._drawing import render_tree

    from ..core.output import truncate_line

    repository, hash_or_tag = image_spec

    if tree:
        render_tree(repository)
    else:
        latest = repository.images["latest"]

        tag_dict = defaultdict(list)
        for img, img_tag in repository.get_all_hashes_tags():
            tag_dict[img].append(img_tag)
        tag_dict[latest.image_hash].append("latest")

        image = repository.images[hash_or_tag]
        log = image.get_log()
        table = []
        for entry in log:
            table.append(
                (
                    format_image_hash(entry.image_hash),
                    format_tags(tag_dict[entry.image_hash]),
                    format_time(entry.created),
                    truncate_line(entry.comment or ""),
                )
            )
        click.echo(tabulate(table, tablefmt="plain"))


def _get_all_tables(repository: "Repository", image_hash: Optional[str]):
    if image_hash is None:
        return repository.engine.get_all_tables(repository.to_schema())
    return repository.images[image_hash].get_tables()


@click.command(name="diff")
@click.option(
    "-v",
    "--verbose",
    default=False,
    is_flag=True,
    help="Include the actual differences rather than just the total number of updated rows.",
)
@click.option("-t", "--table-name", help="Show the differences for a single table.")
@click.argument("repository", type=RepositoryType())
@click.argument("tag_or_hash_1", required=False)
@click.argument("tag_or_hash_2", required=False)
def diff_c(verbose, table_name, repository, tag_or_hash_1, tag_or_hash_2):
    """
    Show differences between two Splitgraph images.

    The two images must be in the same repository. The actual targets of this command depend
    on the number of arguments passed:

    ``sgr diff REPOSITORY``

    Return the differences between the current HEAD image and the checked out schema.

    ``sgr diff REPOSITORY TAG_OR_HASH``

    Return the differences between the image and its parent.

    ``sgr diff REPOSITORY TAG_OR_HASH_1 TAG_OR_HASH_2``

    Return the differences from the first (earlier) image to the second image.
    """
    tag_or_hash_1, tag_or_hash_2 = _get_actual_hashes(repository, tag_or_hash_1, tag_or_hash_2)

    if table_name:
        all_tables = [table_name]
    else:
        all_tables = sorted(
            set(_get_all_tables(repository, tag_or_hash_1)).union(
                set(_get_all_tables(repository, tag_or_hash_2))
            )
        )

    diffs: Dict[str, Union[bool, Tuple[int, int, int], List[Tuple[bool, Tuple]]]] = {
        table_name: repository.diff(table_name, tag_or_hash_1, tag_or_hash_2, aggregate=not verbose)
        for table_name in all_tables
    }

    if tag_or_hash_2 is None:
        click.echo("Between %s and the current working copy: " % tag_or_hash_1[:12])
    else:
        click.echo("Between %s and %s: " % (tag_or_hash_1[:12], tag_or_hash_2[:12]))

    for table, diff_result in diffs.items():
        _emit_table_diff(table, diff_result, verbose)


def _emit_table_diff(
    table_name: str,
    diff_result: Union[bool, Tuple[int, int, int], List[Tuple[bool, Tuple]], None],
    verbose: bool,
) -> None:
    from ..core.output import pluralise

    to_print = "%s: " % table_name
    if isinstance(diff_result, (list, tuple)):
        if verbose:
            change_count = dict(
                Counter(d[0] for d in cast(List[Tuple[bool, Tuple]], diff_result)).most_common()
            )
            added = change_count.get(True, 0)
            removed = change_count.get(False, 0)
            updated = 0
        else:
            added, removed, updated = cast(Tuple[int, int, int], diff_result)

        count = []
        if added:
            count.append("added " + pluralise("row", added))
        if removed:
            count.append("removed " + pluralise("row", removed))
        if updated:
            count.append("updated " + pluralise("row", updated))
        if added + removed + updated == 0:
            count = ["no changes"]
        click.echo(to_print + ", ".join(count) + ".")

        if verbose:
            for added, row in cast(List[Tuple[bool, Tuple]], diff_result):
                click.echo(("+" if added else "-") + " %r" % (row,))
    elif diff_result is None:
        click.echo(to_print + "untracked")
    else:
        # Whole table was either added or removed
        click.echo(to_print + ("table added" if diff_result else "table removed"))


def _get_actual_hashes(
    repository: "Repository", image_1: Optional[str], image_2: Optional[str]
) -> Tuple[str, Optional[str]]:
    if image_1 is None and image_2 is None:
        # Comparing current working copy against the last commit
        image_1 = repository.head_strict.image_hash
    elif image_1 is not None and image_2 is None:
        image_1_obj = repository.images[image_1]
        image_1 = image_1_obj.image_hash
        # One parameter: diff from that and its parent.
        image_2 = image_1_obj.parent_id
        if image_2 is None:
            raise ValueError("%s has no parent to compare to!" % image_1)
        image_1, image_2 = image_2, image_1  # snap_1 has to come first
    else:
        assert image_1 is not None
        assert image_2 is not None
        image_1 = repository.images[image_1].image_hash
        image_2 = repository.images[image_2].image_hash
    return image_1, image_2


@click.command(name="show")
@remote_switch_option()
@click.argument("image_spec", type=ImageType(default="HEAD", get_image=True))
def show_c(image_spec):
    """
    Show information about a Splitgraph image. This includes its parent, comment, size and creation time.

    Note that the size isn't the on-disk footprint, as the image might share some objects with other images
    or if some of the image's objects have not been downloaded.

    Image spec must be of the format ``[NAMESPACE/]REPOSITORY[:HASH_OR_TAG]``. If no tag is specified, ``HEAD`` is used.
    """
    from ..core.output import pretty_size

    repository, image = image_spec

    click.echo("Image %s:%s" % (repository.to_schema(), image.image_hash))
    click.echo(image.comment or "")
    click.echo("Created at %s" % image.created.isoformat())
    click.echo("Size: %s" % pretty_size(image.get_size()))
    if image.parent_id:
        click.echo("Parent: %s" % image.parent_id)
    else:
        click.echo("No parent (root image)")
    click.echo()
    click.echo("Tables:")
    for table in image.get_tables():
        click.echo("  %s" % table)


@click.command(name="table")
@remote_switch_option()
@click.argument("image_spec", type=ImageType(default="HEAD", get_image=True))
@click.argument("table_name", type=str)
@click.option(
    "-v", "--verbose", type=bool, is_flag=True, default=False, help="Show all of table's objects."
)
def table_c(image_spec, table_name, verbose):
    """
    Show information about a table in a Splitgraph image.

    Image spec must be of the format ``[NAMESPACE/]REPOSITORY[:HASH_OR_TAG]``. If no tag is specified, ``HEAD`` is used.
    """
    from ..core.output import pretty_size

    repository, image = image_spec
    table = image.get_table(table_name)
    click.echo("Table %s:%s/%s" % (repository.to_schema(), image.image_hash, table_name))
    click.echo()
    click.echo("Size: %s" % pretty_size(table.get_size()))
    click.echo("Rows: %d" % table.get_length())
    click.echo("Columns: ")
    for column in table.table_schema:
        column_str = "  " + column.name + " (" + column.pg_type
        if column.is_pk:
            column_str += ", PK"
        column_str += ")"
        if column.comment:
            column_str += ": " + column.comment
        click.echo(column_str)

    click.echo()
    click.echo("Objects: ")
    if len(table.objects) > 10 and not verbose:
        click.echo("  " + "\n  ".join(table.objects[:10] + ["..."]))
    else:
        click.echo("  " + "\n  ".join(table.objects))


@click.command(name="object")
@remote_switch_option()
@click.argument("object_id", type=str)
def object_c(object_id):
    """
    Show information about a Splitgraph object.

    Objects, or fragments, are building blocks of Splitgraph tables: each table consists of multiple immutable fragments
    that can partially overwrite each other. Each fragment might have a parent that it depends on. In addition,
    the smallest and largest values for every column are stored in the fragment's metadata. This information is used
    to choose which objects to download in order to execute a query against a table.
    """
    from splitgraph.core.indexing.bloom import describe
    from splitgraph.core.object_manager import ObjectManager
    from splitgraph.engine import ResultShape, get_engine

    from ..core.output import pretty_size
    from ..core.sql import select

    object_manager = ObjectManager(get_engine())
    object_meta = object_manager.get_object_meta([object_id])
    if not object_meta:
        raise click.BadParameter("Object %s does not exist!" % object_id)

    sg_object = object_meta[object_id]
    click.echo("Object ID: %s" % object_id)
    click.echo()
    click.echo("Namespace: %s" % sg_object.namespace)
    click.echo("Format: %s" % sg_object.format)
    click.echo("Size: %s" % pretty_size(sg_object.size))
    click.echo("Created: %s" % sg_object.created)
    click.echo("Rows inserted: %s" % sg_object.rows_inserted)
    click.echo("Insertion hash: %s" % sg_object.insertion_hash)
    click.echo("Rows deleted: %s" % sg_object.rows_deleted)
    click.echo("Deletion hash: %s" % sg_object.deletion_hash)
    click.echo("Column index:")
    for col_name, col_range in sg_object.object_index["range"].items():
        click.echo("  %s: [%r, %r]" % (col_name, col_range[0], col_range[1]))
    if "bloom" in sg_object.object_index:
        click.echo("Bloom index: ")
        for col_name, col_bloom in sg_object.object_index["bloom"].items():
            click.echo("  %s: %s" % (col_name, describe(col_bloom)))

    if object_manager.object_engine.registry:
        # Don't try to figure out the object's location if we're talking
        # to the registry.
        return

    click.echo()
    object_in_cache = object_manager.object_engine.run_sql(
        select("object_cache_status", "1", "object_id = %s"),
        (object_id,),
        return_shape=ResultShape.ONE_ONE,
    )
    object_downloaded = object_id in object_manager.get_downloaded_objects(limit_to=[object_id])
    object_external = object_manager.get_external_object_locations([object_id])

    if object_downloaded and not object_in_cache:
        click.echo("Location: created locally")
    else:
        original_location = (
            ("%s (%s)" % (object_external[0][1], object_external[0][2]))
            if object_external
            else "remote engine"
        )
        if object_in_cache:
            click.echo("Location: cached locally")
            click.echo("Original location: " + original_location)
        else:
            click.echo("Location: " + original_location)


@click.command(name="objects")
@click.option(
    "--local", is_flag=True, help="Show only objects that are physically present on this engine"
)
def objects_c(local):
    """
    List objects known to this engine.
    """
    from splitgraph.core.object_manager import ObjectManager
    from splitgraph.engine import get_engine

    om = ObjectManager(get_engine())

    if local:
        objects = om.get_downloaded_objects()
    else:
        objects = om.get_all_objects()

    click.echo("\n".join(sorted(objects)))


@click.command(name="sql")
@click.argument("sql")
@remote_switch_option()
@click.option("-s", "--schema", help="Run SQL against this schema.")
@click.option(
    "-i",
    "--image",
    help="Run SQL against this image.",
    type=ImageType(default="latest", get_image=True),
)
@click.option("-a", "--show-all", is_flag=True, help="Return all results of the query.")
@click.option("-j", "--json", is_flag=True, help="Return results as JSON")
@click.option("-n", "--no-transaction", is_flag=True, help="Don't wrap the SQL in a transaction.")
def sql_c(sql, schema, image, show_all, json, no_transaction):
    """
    Run an SQL statement against the Splitgraph engine.

    There are no restrictions on the contents of the statement: this is the same as running it
    from any other PostgreSQL client.

    If ``--schema`` is specified, the statement is run with the ``search_path`` set to that schema. This means
    that these statements are equivalent:

    \b
    ```
    sgr sql "SELECT * FROM \"noaa/climate\".table"
    sgr sql -s noaa/climate "SELECT * FROM table"
    ```

    If `--image` is specified, this will run the statement against that image using layered querying.
    Only read-only statements are supported. For example:

    \b
    ```
    sgr sql -i noaa/climate:latest "SELECT * FROM table"
    ```
    """
    from splitgraph.engine import get_engine

    if schema and image:
        raise click.UsageError("Only one of --schema and --image can be specified!")

    engine = get_engine()
    if no_transaction:
        engine.autocommit = True

    if not image:
        if schema:
            engine.run_sql("SET search_path TO %s", (schema,))
        results = engine.run_sql(sql)
    else:
        repo, image = image
        with image.query_schema() as s:
            results = engine.run_sql_in(s, sql)

    emit_sql_results(results, use_json=json, show_all=show_all)


def _emit_repository_data(repositories, engine):
    from tabulate import tabulate

    from splitgraph.engine import ResultShape

    from ..core.output import pretty_size

    click.echo("Local repositories: \n")

    table = []
    for repo, head in repositories:
        no_images = len(repo.images())
        all_tags = repo.get_all_hashes_tags()
        no_tags = len([i for i, t in all_tags if t != "HEAD"])
        metadata_size = pretty_size(repo.get_size())
        actual_size = pretty_size(repo.get_local_size())
        if head:
            head = head.image_hash[:10]
            if (
                engine.run_sql(
                    "SELECT 1 FROM pg_foreign_server WHERE srvname = %s",
                    (("%s_lq_checkout_server" % repo.to_schema())[:63],),
                    return_shape=ResultShape.ONE_ONE,
                )
                is not None
            ):
                head += " (LQ)"
        else:
            head = "--"

        upstream = repo.upstream
        if upstream:
            upstream_text = "%s (%s)" % (upstream.to_schema(), upstream.engine.name)
        else:
            upstream_text = "--"

        table.append(
            (repo.to_schema(), no_images, no_tags, metadata_size, actual_size, head, upstream_text)
        )

    click.echo(
        tabulate(
            table,
            headers=[
                "Repository",
                "Images",
                "Tags",
                "Size (T)",
                "Size (A)",
                "Checkout",
                "Upstream",
            ],
        )
    )
    click.echo("\nUse sgr status REPOSITORY to get information about a given repository.")
    click.echo("Use sgr show REPOSITORY:[HASH_OR_TAG] to get information about a given image.")


@click.command(name="status")
@click.argument("repository", required=False, type=RepositoryType(exists=True))
def status_c(repository):
    """
    Show the status of the Splitgraph engine.

    If a repository is passed, show in-depth information about a repository.

    If not, show information about all repositories local to the engine. This will show a list
    of all repositories, number of local images and tags, total repository size (theoretical
    maximum size and current on-disk footprint of cached objects) and the current checked
    out image (with LQ if the image is checked out using read-only layered querying).
    """
    from splitgraph.core.engine import get_current_repositories
    from splitgraph.engine import get_engine

    if repository is None:
        engine = get_engine()
        repositories = get_current_repositories(engine)
        _emit_repository_data(repositories, engine)
    else:
        head = repository.head
        if not head:
            click.echo("%s: nothing checked out." % str(repository))
            return
        parent, children = head.get_parent_children()
        click.echo("%s: on image %s." % (str(repository), head.image_hash))
        if parent is not None:
            click.echo("Parent: %s" % parent)
        if len(children) > 1:
            click.echo("Children: " + "\n".join(children))
        elif len(children) == 1:
            click.echo("Child: %s" % children[0])



import io
import os
import re
import time
from datetime import date, datetime
from typing import Any, Dict, List, Optional, Union


def pretty_size(size: Union[int, float]) -> str:
    """
    Converts a size in bytes to its string representation (e.g. 1024 -> 1KiB)
    :param size: Size in bytes
    """
    size = float(size)
    power = 2 ** 10
    base = 0
    while size > power:
        size /= power
        base += 1

    return "%.2f %s" % (size, {0: "", 1: "Ki", 2: "Mi", 3: "Gi", 4: "Ti"}[base] + "B")


def pluralise(word: str, number: int) -> str:
    """1 banana, 2 bananas"""
    if word.endswith("y"):
        return "%d %s" % (number, word if number == 1 else word[:-1] + "ies")
    return "%d %s%s" % (number, word, "" if number == 1 else "s")


def truncate_line(line: str, length: int = 80) -> str:
    """Truncates a line to a given length, replacing the remainder with ..."""
    return (line if len(line) <= length else line[: length - 3] + "...").replace("\n", "")


def truncate_list(items: List[Any], max_entries: int = 10) -> str:
    """Print a list, possibly truncating it to the specified number of entries"""
    return ",".join(str(i) for i in items[:max_entries]) + (
        ", ..." if len(items) > max_entries else ""
    )


_slugify = re.compile(r"[^\sa-zA-Z0-9]")


def slugify(text: str, max_length: int = 50) -> str:
    text = _slugify.sub("", text.lower()).strip()
    parts = re.split(r"\s+", text)
    result = parts[0]
    for p in parts[1:]:
        new = result + "_" + p
        if len(new) > max_length:
            break
        result = new
    return result[:max_length]


def parse_repo_tag_or_hash(value, default="latest"):
    repo_image = value.split(":")
    tag_or_hash: Optional[str]
    if len(repo_image) == 2:
        tag_or_hash = repo_image[1]
    else:
        tag_or_hash = default
    from splitgraph.core.repository import Repository

    repo = Repository.from_schema(repo_image[0])
    return repo, tag_or_hash


def conn_string_to_dict(connection: Optional[str]) -> Dict[str, Any]:
    if connection:
        match = re.match(r"((\S+):(\S+)@)?(.+):(\d+)", connection)
        if not match:
            raise ValueError("Invalid connection string!")
        # In the future, we could turn all of these options into actual Click options,
        # but then we'd also have to parse the docstring deeper to find out the types the function
        # requires, how to serialize them etc etc. Idea for a click-contrib addon perhaps?
        result = {"host": match.group(4), "port": int(match.group(5))}
        if match.group(2):
            result["username"] = match.group(2)
        if match.group(3):
            result["password"] = match.group(3)
        return result
    else:
        return {}


def parse_dt(string: str) -> datetime:
    _formats = [
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%f",
        "%Y-%m-%d %H:%M:%S.%f",
    ]
    for fmt in _formats:
        try:
            return datetime.strptime(string, fmt)
        except ValueError:
            continue

    raise ValueError("Unknown datetime format for string %s!" % string)


def parse_date(string: str) -> date:
    return datetime.strptime(string, "%Y-%m-%d").date()


def parse_time(string: str) -> time.struct_time:
    _formats = [
        "%H:%M:%S",
        "%H:%M:%S.%f",
    ]
    for fmt in _formats:
        try:
            return time.strptime(string, fmt)
        except ValueError:
            continue

    raise ValueError("Unknown time format for string %s!" % string)


class ResettableStream(io.RawIOBase):
    """Stream that supports reading from the underlying stream and resetting the position once.

    We can't use fseek() in this case, since we might be reading from a pipe. So, we operate
    this stream in two modes. In the first mode, we mirror all reads into a separate buffer
    (consuming the input stream). After the user calls reset(), we first output data from the
    mirrored copy, then continue consuming the input stream (simulating seek(0).
    """

    def __init__(self, stream):
        self._stream = stream
        self._buffer: Optional[io.BytesIO] = io.BytesIO()
        self._is_reset = False

    def reset(self):
        if self._is_reset:
            raise ValueError("Stream can only be reset once!")
        self._is_reset = True

    def readable(self):
        return True

    def _append_to_buf(self, contents):
        assert self._buffer
        oldpos = self._buffer.tell()
        self._buffer.seek(0, os.SEEK_END)
        self._buffer.write(contents)
        self._buffer.seek(oldpos)

    def readinto(self, b):
        buffer_length = len(b)
        contents = b""
        if self._is_reset and self._buffer:
            assert self._buffer
            # Try reading from the buffer, if it's not exhausted
            contents = self._buffer.read(buffer_length)
            if len(contents) == 0:
                self._buffer = None

        if not contents:
            # Read from the underlying stream
            contents = self._stream.read(buffer_length)

        # If we haven't reset yet, mirror the contents into the buffer
        if not self._is_reset:
            self._append_to_buf(contents)

        b[: len(contents)] = contents
        return len(contents)


class Color:
    """
    An enumeration of console colors
    """

    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"



import csv
import io
import json
import logging
from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Tuple

from minio import Minio

if TYPE_CHECKING:
    import _csv

import chardet

from splitgraph.core.output import ResettableStream

try:
    from multicorn.utils import log_to_postgres
except ImportError:

    def log_to_postgres(*args, **kwargs):
        print(*args)


def dump_options(options: Dict[str, Any]) -> Dict[str, str]:
    return {k: json.dumps(v) for k, v in options.items()}


def load_options(options: Dict[str, str]) -> Dict[str, Any]:
    return {k: json.loads(v) for k, v in options.items()}


class CSVOptions(NamedTuple):
    autodetect_header: bool = True
    autodetect_dialect: bool = True
    autodetect_encoding: bool = True
    autodetect_sample_size: int = 65536
    schema_inference_rows: int = 100000
    delimiter: str = ","
    quotechar: str = '"'
    header: bool = True
    encoding: str = "utf-8"
    ignore_decode_errors: bool = False

    @classmethod
    def from_fdw_options(cls, fdw_options):
        return cls(
            autodetect_header=fdw_options.get("autodetect_header", True),
            autodetect_dialect=fdw_options.get("autodetect_dialect", True),
            autodetect_encoding=fdw_options.get("autodetect_encoding", True),
            autodetect_sample_size=int(fdw_options.get("autodetect_sample_size", 65536)),
            schema_inference_rows=int(fdw_options.get("schema_inference_rows", 100000)),
            header=fdw_options.get("header", True),
            delimiter=fdw_options.get("delimiter", ","),
            quotechar=fdw_options.get("quotechar", '"'),
            encoding=fdw_options.get("encoding", "utf-8"),
            ignore_decode_errors=fdw_options.get("ignore_decode_errors", False),
        )

    def to_csv_kwargs(self):
        return {"delimiter": self.delimiter, "quotechar": self.quotechar}

    def to_table_options(self):
        """
        Turn this into a dict of table options that can be plugged back into CSVDataSource.
        """

        # The purpose is to return to the user the CSV dialect options that we inferred
        # so that they can freeze them in the table options (instead of rescanning the CSV
        # on every mount) + iterate on them.

        # We flip the autodetect flags to False here so that if we merge the new params with
        # the old params again, it won't rerun CSV dialect detection.
        return {
            "autodetect_header": False,
            "autodetect_dialect": False,
            "autodetect_encoding": False,
            "header": self.header,
            "delimiter": self.delimiter,
            "quotechar": self.quotechar,
            "encoding": self.encoding,
        }


def autodetect_csv(stream: io.RawIOBase, csv_options: CSVOptions) -> CSVOptions:
    """Autodetect the CSV dialect, encoding, header etc."""
    if not (
        csv_options.autodetect_encoding
        or csv_options.autodetect_header
        or csv_options.autodetect_dialect
    ):
        return csv_options

    data = stream.read(csv_options.autodetect_sample_size)
    assert data

    if csv_options.autodetect_encoding:
        encoding = chardet.detect(data)["encoding"]
        if encoding == "ascii" or encoding is None:
            # ASCII is a subset of UTF-8. For safety, if chardet detected
            # the encoding as ASCII, use UTF-8 (a valid ASCII file is a valid UTF-8 file,
            # but not vice versa)

            # If we can't detect the encoding, fall back to utf-8 too (hopefully the user
            # passed ignore_decode_errors=True
            encoding = "utf-8"
        csv_options = csv_options._replace(encoding=encoding)

    sample = data.decode(
        csv_options.encoding, errors="ignore" if csv_options.ignore_decode_errors else "strict"
    )
    # Emulate universal newlines mode (convert \r, \r\n, \n into \n)
    sample = "\n".join(sample.splitlines())

    if csv_options.autodetect_dialect:
        dialect = csv.Sniffer().sniff(sample)
        # These are meant to be set, but mypy claims they might not be.
        csv_options = csv_options._replace(
            delimiter=dialect.delimiter or ",", quotechar=dialect.quotechar or '"'
        )

    if csv_options.autodetect_header:
        has_header = csv.Sniffer().has_header(sample)
        csv_options = csv_options._replace(header=has_header)

    return csv_options


def make_csv_reader(
    response: io.IOBase, csv_options: CSVOptions
) -> Tuple[CSVOptions, "_csv._reader"]:
    stream = ResettableStream(response)
    csv_options = autodetect_csv(stream, csv_options)

    stream.reset()
    # https://docs.python.org/3/library/csv.html#id3
    # Open with newline="" for universal newlines
    io_stream = io.TextIOWrapper(
        io.BufferedReader(stream),
        encoding=csv_options.encoding,
        newline="",
        errors="ignore" if csv_options.ignore_decode_errors else "strict",
    )

    reader = csv.reader(io_stream, **csv_options.to_csv_kwargs(), skipinitialspace=True)
    return csv_options, reader


def get_s3_params(fdw_options: Dict[str, Any]) -> Tuple[Minio, str, str]:
    s3_client = Minio(
        endpoint=fdw_options["s3_endpoint"],
        access_key=fdw_options.get("s3_access_key"),
        secret_key=fdw_options.get("s3_secret_key"),
        secure=fdw_options.get("s3_secure", True),
        region=fdw_options.get("s3_region"),
    )

    s3_bucket = fdw_options["s3_bucket"]

    # We split the object into a prefix + object ID to let us mount a bunch of objects
    # with the same prefix as CSV files.
    s3_object_prefix = fdw_options.get("s3_object_prefix", "")

    return s3_client, s3_bucket, s3_object_prefix


def pad_csv_row(row: List[str], num_cols: int, row_number: int) -> List[str]:
    """Preprocess a CSV file row to make the parser more robust."""

    # Truncate/pad the row to the expected number of columns to match the header. We'd
    # rather return a CSV file full of varchars and NaNs than error out directly.
    row_len = len(row)
    if row_len > num_cols:
        log_to_postgres(
            "Row %d has %d column(s), truncating" % (row_number, row_len), level=logging.DEBUG
        )
        row = row[:num_cols]
    elif row_len < num_cols:
        log_to_postgres(
            "Row %d has %d column(s), padding" % (row_number, row_len), level=logging.DEBUG
        )
        row.extend([""] * (num_cols - row_len))
    return row



"""
Splitfile interpreter: a set of tools on top of the core Splitgraph versioning and image management to give
the user a Dockerfile-like experience for building Splitgraph images (caching, consistent hashing, a declarative
language).
"""
